@@ -55,8 +55,14 @@ let call = match post with
   end ;
   call#set_request_header http_headers ;
   pipeline#add call ;
-  pipeline#run () ;
-  call#response_body#value
+  try
+    pipeline#run () ;
+    call#response_body#value
+  with
+    | Http_client.Http_protocol e 
+    | e -> 
+       pipeline#reset() ; 
+       raise (Http  (Printexc.to_string e))
 
 module Audioscrobbler=
   struct
@@ -67,6 +73,7 @@ module Audioscrobbler=
     (* Data types *)
     type source = User | Broadcast | Recommendation | Lastfm | Unknown
     type rating = Love | Ban | Skip
+    type action = NowPlaying | Submit
 
     (* song submission type *)
     type song = { artist : string; track: string; time: float option; 
@@ -77,7 +84,7 @@ module Audioscrobbler=
 
     type error = Http of string | Banned | Badauth | Badtime
                 | Failed of string | UnknownError of string | Success
-                | Internal of string
+                | Internal of string | BadData of string
     exception Error of error
     
     let string_of_error e = 
@@ -89,6 +96,7 @@ module Audioscrobbler=
         | Failed s -> Printf.sprintf "failure: %s" s
         | UnknownError s -> Printf.sprintf "unknown error: %s" s
         | Internal s -> Printf.sprintf  "erreur interne: %s" s
+        | BadData s -> Printf.sprintf "Wrong song data: %s" s
         | Success -> "success!"
 
     let error_of_response s = 
@@ -179,6 +187,30 @@ module Audioscrobbler=
                           Hashtbl.replace urls id v; id
              | e -> raise e
 
+    let check_song s a = 
+      match s.source with
+        | None when a = Submit ->
+             raise (BadData "source field is required for submit action")
+        | _ -> () ;
+      match s.length with
+        | None when a = Submit && s.source = Some User -> 
+            raise (BadData "length required when source is User")
+        | _ -> () ;
+      match s.time with
+        | None when s.source = Some User ->
+             raise (BadData "time required when source is User")
+        | _ -> () ;
+      match s.trackauth with
+        | None when s.source = Some Lastfm -> 
+             raise (BadData "trackauth required when source is Lastfm")
+        | _ -> () ;
+      match s.rating with 
+        | Some Ban
+        | Some Skip when s.source <> Some Lastfm -> 
+             raise (BadData "bad rating value (ban and skip are for lastfm sources only)")
+        | _ -> ()
+
+
     let audioscrobbler_post id base_url values = 
       let url = Neturl.parse_url base_url in
       let host = Neturl.url_host url in
@@ -201,8 +233,9 @@ module Audioscrobbler=
         try 
           Hashtbl.find urls id
         with
-          | Not_found -> raise (Internal (Printf.sprintf "No session data for session ID %s" id))
+          | Not_found -> raise Badauth
       in
+      check_song song NowPlaying ;
       let values = [("s",id);("a",song.artist);("t",song.track);
                     ("album",arg_value_string song.album);
                     ("l",arg_value_float song.length);
@@ -219,61 +252,84 @@ module Audioscrobbler=
           | Not_found -> raise (Internal (Printf.sprintf "No session data for session ID %s" id))
       in
       let count = ref 0 in
+      let failed = ref [] in
       let args = ref [("s",id)] in
       let get_arg = fun s -> Printf.sprintf "%s[%d]" s !count in
       let add_arg song =
-        let ar,tr,t,s,r,l,al,n,m,x =
-           song.artist, song.track,song.time,
-           song.source, song.rating,
-           song.length, song.album,
-           song.tracknumber, song.musicbrainzid,
-           song.trackauth
-        in
-        let l = match l with
-                  | None when s <> Some User -> ""
-                  | Some s -> Printf.sprintf "%.0f" s
-                  | None -> raise (Internal "length required when source is User")
-        in
-        let t = match t with
-                  | Some t -> Printf.sprintf "%.0f" t
-                  | None -> raise (Internal "time required when source is User")
-        in
-        let x = match x with
-                  | Some x -> x
-                  | None when s = Some Lastfm -> raise (Internal "trackauth required when source is Lastfm")
-                  | None -> ""
-        in
-        let s =  match s with 
-                   | Some User -> "P"
-                   | Some Broadcast -> "P"
-                   | Some Recommendation -> "E"
-                   | Some Lastfm -> "L"
-                   | Some Unknown -> "U"
-                   | None -> raise (Internal "source field is required for submit")
-        in
-        let r =  match r with 
-                   | Some Love -> "L"
-                   | Some Ban when s = "L" -> "B"
-                   | Some Skip when s = "L" -> "S"
-                   | None -> ""
-                   | _ -> raise (Internal "bad rating value (ban and skip are for lastfm sources only)")
-        in
-        args := [(get_arg "a",ar);(get_arg "t",tr);(get_arg "i",t);
-                 (get_arg "o",s ^ x);(get_arg "r",r);(get_arg "l",l);
-                 (get_arg "b",arg_value_string al);(get_arg "n",arg_value_int n);
-                 (get_arg "m",arg_value_string m)] @ (!args)
+        try 
+          check_song song Submit ;
+          let ar,tr,t,s,r,l,al,n,m,x =
+             song.artist, song.track,song.time,
+             song.source, song.rating,
+             song.length, song.album,
+             song.tracknumber, song.musicbrainzid,
+             song.trackauth
+          in
+          let l = match l with
+                    | None -> ""
+                    | Some s -> Printf.sprintf "%.0f" s
+          in
+          let t = match t with
+                    | Some t -> Printf.sprintf "%.0f" t
+                    | None -> ""
+          in
+          let x = match x with
+                    | Some x -> x
+                    | None -> ""
+          in
+          let s =  match s with 
+                     | Some User -> "P"
+                     | Some Broadcast -> "P"
+                     | Some Recommendation -> "E"
+                     | Some Lastfm -> "L"
+                     | Some Unknown -> "U"
+                     | None -> raise (Internal "source field is required for submit")
+          in
+          let r =  match r with 
+                     | Some Love -> "L"
+                     | Some Ban -> "B"
+                     | Some Skip -> "S"
+                     | None -> ""
+          in
+          args := [(get_arg "a",ar);(get_arg "t",tr);(get_arg "i",t);
+                   (get_arg "o",s ^ x);(get_arg "r",r);(get_arg "l",l);
+                   (get_arg "b",arg_value_string al);(get_arg "n",arg_value_int n);
+                   (get_arg "m",arg_value_string m)] @ (!args)
+       with
+         | Error e -> failed := (e,song) :: !failed 
      in
      List.iter add_arg songs ;
-     audioscrobbler_post id url (!args)
+     audioscrobbler_post id url (!args) ;
+     !failed
+
+    let get_song ?time ?source ?rating
+                 ?length ?album ?tracknumber 
+		 ?musicbrainzid ?trackauth 
+		 ~artist ~track () = 
+    { artist = artist ; track = track ; time = time  ;
+      source = source; rating = rating ; length = length ;
+      album = album ; tracknumber = tracknumber ; 
+      musicbrainzid = musicbrainzid ; trackauth = trackauth }
 
     let do_np client login song = 
       let id = handshake client login in
-      np id song
+      try
+        np id song
+      with
+        | Error Badauth -> (* Retry in case of expired session id *)
+	                   clear id ;
+	                   let id = handshake client login in
+		           np id song
 
     let do_submit client login songs = 
       let id = handshake client login in
-      submit id songs
-
+      try 
+        submit id songs
+      with
+        | Error Badauth -> (* Retry in case of expired session id *)
+                           clear id ;
+                           let id = handshake client login in
+                           submit id songs 
 
   end
 
@@ -285,57 +341,95 @@ module Radio=
      * and an uri *)
     type track = (string * string) list * string
 
-    type error = Http of string | Init of string | Adjust of string*string | Playlist | Empty
+    type session = { login : login option; station : ((string*string) list*string) option;
+                     base_url : string; playlist_url : string option; 
+                     base_path: string }
+
+    type error = Http of string | Auth of string | Adjust of string*string | Playlist | Empty
     exception Error of error
-    
+    exception Internal of string
+
     let string_of_error e = 
       match e with
         | Http s -> Printf.sprintf "http connection failed: %s" s
-        | Init s -> Printf.sprintf "could not open session:\n%s" s
+        | Auth s -> Printf.sprintf "could not open session:\n%s" s
         | Adjust (s,s') -> Printf.sprintf "could not adjust station to %s:\n%s\nIs the URI valid ?" s s'
-        | Playlist -> "error while parsinf the playlist"
+        | Playlist -> "error while parsing the playlist"
         | Empty -> "no files available"
-    
+   
+    let _raise = raise
     let raise e = raise (Error e)
+    (* Wrapper for main request *)
+    let request ~host = try request ~host:host with e -> raise (Http (Printexc.to_string e))
 
     (* Some constant for the protocol *)
-    let host = "ws.audioscrobbler.com"
-    let port = 80
-    let sessions = Hashtbl.create 1
-    let stations = Hashtbl.create 1
-    let anon_req = "//1.0/webclient/xmlrpc.php"
+    let base_host = "ext.last.fm"
+    let sessions = Hashtbl.create 10
+    let anon_req = "/1.0/webclient/xmlrpc.php"
     let anon_post = "<methodCall><methodName>getSession</methodName><params /></methodCall>"
     let anon_handshake = Printf.sprintf "/1.0/radio/webclient/handshake.php?sessionKey=%s&user=%s"
     let registered_handshake = Printf.sprintf "/radio/handshake.php?username=%s&passwordmd5=%s"
-    let station_set = Printf.sprintf "/radio/adjust.php?session=%s&url=%s"
-    let anon_split_rex = Pcre.regexp "<value><string>(.+)</string></value>\n<value><string>(.+)</string></value>"
+    let station_set base id url = Printf.sprintf "%s/adjust.php?session=%s&url=%s" base id url
 
-    let playlist_req id opt = 
-      let opt = 
-        match opt with 
-          | None -> ""
-          | Some s -> Printf.sprintf "&%s" s
+    let playlist_req id options = 
+      let d = 
+        try
+         Hashtbl.find sessions id
+        with
+          | Not_found -> 
+              raise (Auth (Printf.sprintf "No session registered for id %s" id))
       in
-      Printf.sprintf "/radio/xspf.php?sk=%s%s&desktop=1" id opt
-
-    (* Wrapper for main request *)
-    let request = try request ~port:port ~host:host with e -> raise (Http (Printexc.to_string e))
+      let options = 
+        match options with
+          | Some s -> s
+          | None -> ""
+      in
+      let url = 
+        match d.playlist_url with
+          | None -> "http://ws.audioscrobbler.com/radio/xspf.php"
+          | Some s -> s
+      in
+      Printf.sprintf "%s?sk=%s&%s" url id options
 
+    let playlist id options = 
+       let url = playlist_req id options in
+       let data = 
+         try
+           Http_client.Convenience.http_get url
+         with
+           | e -> raise (Http (Printexc.to_string e))
+       in
+       let data = Netencoding.Base64.decode data in
+       Netencoding.Url.decode data
     
     (* Some parsing functions *)
-    
-    let handshake_rex = Pcre.regexp "session=([0-9a-z]+).*"
-    let parse_handshake s = 
+
+    let parse_args s = 
+      let rex = Pcre.regexp "[&\n]" in
+      let values = Pcre.split ~rex:rex s in
+      let split s l = 
         try
-          let sub = Pcre.exec ~rex:handshake_rex s in
-          Pcre.get_substring sub 1
+          let sub = Pcre.exec ~pat:"([^=]*)=(.*)" s in
+          (Pcre.get_substring sub 1,Pcre.get_substring sub 2) :: l
         with
-          | Not_found -> raise (Init s)
+          | Not_found -> l
+      in
+      List.fold_right split values []
     
+    let parse_handshake s =
+      let params = parse_args s in
+      try
+        let f x = List.assoc x params in
+        let g x = try Some (f x) with Not_found -> None in
+          f "session",g "playlist_url", 
+          f "base_url", f "base_path"
+      with
+        | Not_found -> raise (Auth s)
+
     let adjust_pat = "response=OK"
     let check_adjust s =
         Pcre.pmatch ~pat:adjust_pat s
-
+    
     let opt_split_rex = Pcre.regexp "^([^?]+)\\?(.+)$"
     let opt_parse s =
         try
@@ -343,18 +437,25 @@ module Radio=
           Pcre.get_substring sub 1, Some (Pcre.get_substring sub 2)
         with
           | Not_found -> s,None
-    
+
     let auth_split_rex = Pcre.regexp "^lastfm://([^:]+):([^@]+)@(.+)$"
     let parse uri = 
         try
           let sub = Pcre.exec ~rex:auth_split_rex uri in
+          let data = Pcre.get_substring sub 3 in
+          let station,options = opt_parse data in
           let user,password = Pcre.get_substring sub 1,
                               Pcre.get_substring sub 2
           in
           (Some { user = user ; password = password }),
-                 opt_parse (Pcre.get_substring sub 3)
+                 Printf.sprintf "lastfm://%s" station,options
         with
-          | Not_found -> None,opt_parse uri
+          | Not_found -> let station,options = 
+                           opt_parse uri 
+                         in
+                         None,
+                         station,
+                         options
     
     let anon_parse s = 
        try
@@ -366,85 +467,137 @@ module Radio=
             	Xml.Element("value",_,[Xml.Element("string",_,[Xml.PCData y])])]) :: l
     	          -> x,y
              | Xml.Element(_,_,l) :: l' -> get_auth (l@l')
-    	 | _ -> raise (Init s)
+    	 | _ -> raise (Auth s)
         in
         get_auth [content]
        with
-         | _ -> raise (Init s)
+         | _ -> raise (Auth s)
     
     (* Core stuff.. *)
     
-    let clear sessionid =
-        let keys = Hashtbl.fold (fun a b r -> if b = sessionid then a::r else r)
-                      sessions []
-        in
-        ignore (List.map (fun x -> Hashtbl.remove sessions x) keys) ;
-        Hashtbl.remove stations sessionid
+    let clear id =
+        Hashtbl.remove sessions id
     
     let anon_session () = 
       let headers = [("Content-Type","text/xml")] in
-      let ret = request ~post:anon_post ~headers:headers anon_req in
+      let ret = request ~host:base_host ~post:anon_post ~headers:headers anon_req in
       anon_parse ret
     
     let init login = 
       match login with
         | None -> let user,id = anon_session () in
-                  let ret = request 
+                  let ret = request ~host:base_host 
                       (anon_handshake id (Netencoding.Url.encode user)) 
                   in
-                  parse_handshake ret
+                  let id,playlist_url,
+                      base_url,base_path = 
+                      parse_handshake ret 
+                  in
+		  Hashtbl.replace sessions id {playlist_url=playlist_url;
+                                               base_url=base_url;
+                                               base_path=base_path;
+                                               login=None;
+                                               station=None} ;
+		  id
        | Some login -> 
-          let user,password = login.user,login.password in
           try
-            Hashtbl.find sessions (user,password)
+            Hashtbl.iter (fun x d -> 
+	                     if d.login = Some login then _raise (Internal x))
+			 sessions ;
+	    _raise Not_found
           with
             | Not_found -> 
+	            let user,password = login.user,login.password in
                     let password = Digest.to_hex (Digest.string password) in
-                    let ret = request (registered_handshake 
+                    let ret = request ~host:base_host (registered_handshake 
     	                      (Netencoding.Url.encode user) password)
                     in
-                    let sessionid = parse_handshake ret in
-                    Hashtbl.replace sessions (user,password) sessionid;
-                    sessionid
+                    let id,playlist_url,
+                        base_url,base_path = 
+                        parse_handshake ret 
+                    in
+		    Hashtbl.replace sessions id {playlist_url=playlist_url;
+                                                 base_url=base_url;
+                                                 base_path=base_path;
+                                                 login=Some login;
+                                                 station=None} ;
+		    id
+	   | Internal x -> x
       
     
-    let adjust sessionid req = 
-      try 
-        assert (Hashtbl.find stations sessionid = req)
-      with 
-        | _ ->
-        let http_req = station_set sessionid (Netencoding.Url.encode req)
-        in
-        let ret = request http_req in
-        if check_adjust ret then
-           Hashtbl.replace stations sessionid req
-        else
-           begin
-             Hashtbl.remove stations sessionid; 
-             clear sessionid;
-             raise (Adjust (req,ret))
-           end
+    let adjust id req = 
+      let d = 
+        try
+          Hashtbl.find sessions id
+	with
+	  | Not_found -> 
+	      raise (Auth (Printf.sprintf "No session registered for id %s" id))
+      in
+      let base_url,base_path = d.base_url,d.base_path in
+        match d.station with
+	  | Some (a,s) when s = req -> a
+          | _ ->
+                let http_req = station_set base_path id 
+                   (Netencoding.Url.encode req)
+                in
+                let ret = request ~host:base_url http_req in
+                if check_adjust ret then
+                 let args = parse_args ret in
+                 ( Hashtbl.replace sessions id 
+                        {login=d.login;station=Some (args,req);
+                         playlist_url=d.playlist_url;
+                         base_url=base_url;base_path=base_path} ;
+                  args )
+                else
+                  begin
+                    clear id;
+                    raise (Adjust (req,ret))
+                  end
     
-    let tracks sessionid opt = 
+    let tracks id options = 
       try
-        let req = playlist_req sessionid opt in
-        let playlist = request req in
+        let playlist = playlist id options in
         Xmlplaylist.tracks playlist
       with
-        | Xmlplaylist.Error e -> clear sessionid; raise Playlist 
-        | Error e -> clear sessionid; raise e
+        | Xmlplaylist.Error e -> clear id; raise Playlist 
+        | Error e -> clear id; raise e
     
     let get uri = 
-      let login,(station,options) = parse uri in
-      let id = init login in
-      adjust id station;
-      tracks id options
-    
-    let url id opt = 
-     Printf.sprintf "http://%s:%d%s" host port (playlist_req id opt)
-
-    let playlist id opt = 
-      request (playlist_req id opt)
+      let login,station,options = parse uri in
+      let id = 
+        match login with
+	  | None -> 
+	    (* Check if another anonymous session is registered
+	       for this station url.
+	       
+	       However, this cannot be optimal, we may end up
+	       asking a new playlist for another session, which 
+	       may stop any other simultaneous streaming on the 
+	       same session. *)
+	    begin
+	      try
+	        let ids = 
+                  Hashtbl.fold (fun a d c -> 
+		                  match d.station with
+                                    | Some (_,s) when s = station -> a :: c
+                                    | _ -> c)
+			       sessions []
+		in
+		List.hd ids
+	      with
+	        |  Failure "hd" -> init login
+	    end
+	  | _ -> init login 
+      in
+      try
+        ignore(adjust id station);
+        tracks id options
+      with Error _ when login <> None -> 
+          (* Retry in case of expired session *)
+	  clear id;
+          let id = init login in
+	  ignore(adjust id station);
+	  tracks id options
 
 end
 
